/*
file:			log.cpp
author:		Allan Kelly
date:		10 August 1999
*/
#pragma warning(disable:4786)

// System includes --------------------
#include <map>

// Project includes -------------------
#include <Critical.h>

// Local includes ---------------------
#include "Log.h"

namespace Log
{
	// pointer to a member of LogSink
	typedef void (LogSink::* SinkMemberPtr) (const std::string&) const;

	// this class conatins the log sinks
	// it also guards against mutli-thread corruption
	// as all functions are inline there is no performance hit
	class SinkStore
	{
		typedef std::map<std::string, LogSink*> SinkMap;
		CriticalSection		_critSec;
		SinkMap				_sinks;
	public:
		SinkStore() {}

		~SinkStore()
		{
			CriticalToken token(_critSec);
			while (!_sinks.empty())
			{
				SinkMap::iterator	i = _sinks.begin();
				delete (*i).second;
				_sinks.erase((*i).first);
			}
		}
		
		void AdoptSink(	const std::string& name,
						LogSink *sink)
		{
			CriticalToken token(_critSec);
			_sinks[name] = sink;
		}

		LogSink* OrphanSink(const std::string& name)
		{
			SinkMap::const_iterator i = _sinks.find(name);
			if (_sinks.end() == i)
			{
				return NULL;
			}
			LogSink* s = (*i).second;
			_sinks.erase(i);
			return s;
		}

		void WriteMsg(SinkMemberPtr func, const std::string& msg)
		{
			CriticalToken token(_critSec);
			for (SinkMap::iterator	i = _sinks.begin();
									i != _sinks.end();
									i++)
			{
					LogSink* s = (*i).second;
					(s->*func)(msg);
			}
		}
	};	// SinkStore class

	SinkStore& Sinks()
	{
		// sinks is the single instance the sink store
		// to ensure it is initialised before use wrap it in a function
		static SinkStore sinks;
		return sinks;
	}

	// need and instance of 
	LogStream& MsgOut()
	{
		static LogStream out;
		return out;
	}

	void AdoptSink(	const std::string& name,
					LogSink* const s)
	{
		Sinks().AdoptSink(name, s);
	}

	LogSink* OrphanSink(const std::string& name)
	{
		return Sinks().OrphanSink(name);
	}

	// functions in the namespace ----------------------------
	void Debug(const std::string& msg)
	{
		std::string m("Debug: ");
		m += msg;
		Sinks().WriteMsg(&LogSink::Debug, m);	
	}

	LogStream& operator<<(LogStream& st, DebugMsg& msg)
	{
		msg << '\0';	// terminate the msg quickly
		std::string text(msg.str());
		msg.freeze(false);
		Debug(text);
		return st;
	}

	void Warning(const std::string& msg)
	{
		std::string m("Warning: ");
		m += msg;
		Sinks().WriteMsg(&LogSink::Warning, m);	
	}

	LogStream& operator<<(LogStream& st, WarningMsg& msg)
	{
		msg << '\0';	// terminate the msg quickly
		std::string text(msg.str());
		msg.freeze(false);
		Warning(text);
		return st;
	}

	void Information(const std::string& msg)
	{
		std::string m("Information: ");
		m += msg;
		Sinks().WriteMsg(&LogSink::Information, m);	
	}

	LogStream& operator<<(LogStream& st, InformationMsg& msg)
	{
		msg << '\0';	// terminate the msg quickly
		std::string text(msg.str());
		msg.freeze(false);
		Information(text);
		return st;
	}

	void Error(const std::string& msg)
	{
		std::string m("Error: ");
		m += msg;
		Sinks().WriteMsg(&LogSink::Error, m);	
	}

	LogStream& operator<<(LogStream& st, ErrorMsg& msg)
	{
		msg << '\0';	// terminate the msg quickly
		std::string text(msg.str());
		msg.freeze(false);
		Error(text);
		return st;
	}
}  // namespace
